/**************************************************************************
***    
*** Copyright (c) 1995-2000 Regents of the University of California,
***               Andrew E. Caldwell, Andrew B. Kahng and Igor L. Markov
*** Copyright (c) 2000-2002 Regents of the University of Michigan,
***               Saurabh N. Adya and Igor L. Markov
***
***  Contact author(s): abk@cs.ucsd.edu, imarkov@umich.edu
***  Original Affiliation:   UCLA, Computer Science Department,
***                          Los Angeles, CA 90095-1596 USA
***
***  Permission is hereby granted, free of charge, to any person obtaining 
***  a copy of this software and associated documentation files (the
***  "Software"), to deal in the Software without restriction, including
***  without limitation 
***  the rights to use, copy, modify, merge, publish, distribute, sublicense, 
***  and/or sell copies of the Software, and to permit persons to whom the 
***  Software is furnished to do so, subject to the following conditions:
***
***  The above copyright notice and this permission notice shall be included
***  in all copies or substantial portions of the Software.
***
*** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
*** EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
*** OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
*** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
*** CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
*** OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
*** THE USE OR OTHER DEALINGS IN THE SOFTWARE.
***
***
***************************************************************************/







// 97.8.8  Paul Tucker 
// 970820 pat  added x-coord sort to subRows in RBPlacement
// 970821 pat  initialize subrow spacing
// 970823 ilm  added orientations to RBPCoreRow and RBPIORow (Maogang's request)//                  they are initialized now
// 970902 ilm  RBPlacement::RBPlacement(const DB&) does not populate rows
//                  with cell Id numbers anymore. This is done with
//                  RBPlacement::populateRowsWith(const Placement&)
//                  (the code has to be exteneded to put IOPads intoIORows,
//                  and a lot more)
// 970917 pat  fix bug in RBCPlacement ctor for test creation
// 970918 pat  modify populateRowsWith to only put core cells in
//             core subrows, and use binary search.
//             improve operator<< for subRows.
// 970919 pat  more elaborate operator<< for RBCSubRowCol
// 970921 pat  upgraded constructors to split core subrows 
//             according to special net wiring;
//             added real RBCPlacement ctor
// 970923 pat  finish RBCPlacement; delete dead code; obey object interfaces
// 970930 pat  erase subRow index vectors at start of populateRowsWith,
//             so legalizer can move cells around
// 970930 pat  add findCoreRow and findSubRow search functions
// 971001 pat  moved comparison functors into .h file
// 971129 pat  comment out old diagnostic message about # of
//             cells not placed in core rows
// 980211 pat  add handling for missing cases to construction of
//             RBPSubRows based on special nets
// 982202 pat  corrected error of creating subRows
// 982202  dv  changed the interface of populateRowsWith, added two more
//             unsigned Data Members, added support for isPopulated.
// 980306  dv  removed orientedRBPlacement and added its functionality to
//             RBPlacement
// 980309 pat  more const declarations
// 980427 pat  added findSubRow as a coreRow method.
// 990527 aec  split RBPlacement into a base class + derived class
//			base class does not use DB
// 990609 s.m  added spaceCellsWithPinDimensionsAlg1()
// 990610 s.m  added spaceCellsWithPinDimensionsAlg1()
// 000803 ilm  fixed savePlacement() to save cell names    

#ifdef _MSC_VER
#pragma warning(disable:4786)
#endif

#include <algorithm>
#include <ABKCommon/sgi_hash_map.h>
#include "ABKCommon/abkcommon.h"
#include "RBPlacement.h"

#ifdef _MSC_VER
#ifndef rint
#define rint(a) floor((a)+0.5)
#endif
#endif

void RBPlacement::splitRowByObstacle(RBPCoreRow& cr, const BBox& obstacle)
{
    for(unsigned srIdx = 0; srIdx < cr.getNumSubRows(); srIdx++)
    {
      RBPSubRow& sr = cr._subRows[srIdx];
      double     srXStart = sr.getXStart();
      double     srXEnd   = sr.getXEnd();

      if(obstacle.xMin >= srXEnd || obstacle.xMax <= srXStart)
		continue;

      // There are four ways the obstacle can intersect the subrow in x-dir
      // |----------| 		Obstacle
      //    |---|     		SubRow
      //
      //      |---|     	Obstacle
      //   |----------| 	SubRow
      //
      // |---|          	Obstacle
      //   |----------| 	SubRow
      //
      //            |---| 	Obstacle
      //   |----------|   	SubRow


      if((obstacle.xMin <= srXStart) && (obstacle.xMax >= srXEnd))
      { // Case1 : delete the subrow

        cr._subRows.erase(cr._subRows.begin() + srIdx);
        --srIdx;
        break;
      }
      else if((obstacle.xMin <= srXStart) && 
              (obstacle.xMax >  srXStart) && 
	      (obstacle.xMax <  srXEnd))
      { // Case3: trim the front of the subrow.
		
	double   newXStart    = obstacle.xMax - srXStart;
	unsigned removedSites = 
			static_cast<unsigned>(ceil(newXStart/sr.getSpacing()))+1;
	newXStart = srXStart + removedSites*sr.getSpacing();

        sr._xStart = newXStart;
	sr.resetNumSites();
	if(sr.getNumSites() <= 1)
	{
	    cr._subRows.erase(cr._subRows.begin() + srIdx);
	    srIdx--;
	}
      }
      else if((obstacle.xMin >  srXStart) && 
              (obstacle.xMin <  srXEnd)   && 
              (obstacle.xMax >= srXEnd)   )
      {	// Case4: trim the end of the subrow

        double newLength = (obstacle.xMin-srXStart)-sr.getSiteWidth();
	if(newLength <= sr.getSiteWidth())
	{
	    cr._subRows.erase(cr._subRows.begin() + srIdx);
	    srIdx--;
	}
	else
	{ 
	    newLength -= sr.getSiteWidth();
	    unsigned numSites = 
			static_cast<unsigned>(floor(newLength / sr.getSpacing()));
	    double newXEnd = sr.getXStart() + numSites*sr.getSpacing() +
				sr.getSiteWidth();
            sr._xEnd = newXEnd;
	    sr.resetNumSites();
	    abkassert(sr.getNumSites() == numSites+1, "#sites mismatch");
	    if(sr.getNumSites() == 0)
	    {
		cr._subRows.erase(cr._subRows.begin() + srIdx);
		srIdx--;
	    }
	}
      }
      else if((obstacle.xMin > srXStart) && (obstacle.xMax < srXEnd))
      { 	// Case2: split the subrow in two
	      	// The new subRow will be the second one

        RBPSubRow newSR(sr);
	double newXStart = obstacle.xMax - srXStart;
	unsigned removedSites = 
			static_cast<unsigned>(ceil(newXStart/sr.getSpacing())+1);
	newXStart = srXStart + removedSites*sr.getSpacing();
        newSR._xStart = newXStart;
	newSR.resetNumSites();

	double newLength = (obstacle.xMin-srXStart)-sr.getSiteWidth();
	newLength -= sr.getSiteWidth();
	if(newLength > 0)
	{
	    unsigned numSites = 
			static_cast<unsigned>(floor(newLength / sr.getSpacing()));
	    double newXEnd = srXStart + numSites*sr.getSpacing() +
				sr.getSiteWidth();
            sr._xEnd = newXEnd;
	    sr.resetNumSites();
	    abkassert(sr.getNumSites() == numSites+1, "#sites mismatch");
	}
	else
	    sr._numSites = 0;

	if(sr.getNumSites() <= 1)
	{
	    cr._subRows.erase(cr._subRows.begin() + srIdx);
	    srIdx--;
	}
	if(newSR.getNumSites() > 1)
	    cr._subRows.insert(cr._subRows.begin() + srIdx +1, newSR);
      }
    }
}


void RBPlacement::resetPlacementWOri(const PlacementWOrient& pl)
{
    _placement    = pl;
    setCellsInRows();
    populateCC();
}

void RBPlacement::resetPlacement(const Placement& pl)
{
    _placement    = PlacementWOrient(pl, vector<Orient>(pl.size()));
    setCellsInRows();
    populateCC();
}

void RBPlacement::setCellsInRows()
  // Puts all the cellIds onto their appropriate subRows.  
{
  _populated = true;
  _cellsNotInRows = 0;

  //clear all rows
  std::fill(_isInSubRow.begin(), _isInSubRow.end(), false);

  unsigned i;
  for(i=0; i<_coreRows.size(); i++)
    for(unsigned j=0; j<_coreRows[i].getNumSubRows(); j++)
    {
      RBPSubRow& sRow = _coreRows[i][j];
      sRow.clearAllCells();
    } 

  for(i=0; i<getNumCells(); i++)
  {
    if(_isCoreCell[i])
    {
      RBRowPtrs rowPtrs = findBothRows(_placement[i]);

      if(rowPtrs.first != NULL && rowPtrs.second != NULL)
      {
//        const_cast<RBPSubRow*>(rowPtrs.second)->addCellToEnd(i);
          const_cast<RBPSubRow*>(rowPtrs.second)->addCellToEnd(i);
	  _isInSubRow[i] = true;
      }	
      else
          _cellsNotInRows++;
    }
  }
  if (_cellsNotInRows) 
    cout<<" RBPlacement:"<<_cellsNotInRows<<" core cell(s) not in rows"<<endl;

  sortSubRowsByX();
}

void RBPlacement::populateCC ()
{
    for(unsigned crIdx = 0; crIdx < _coreRows.size(); ++crIdx)
    {
    	const RBPCoreRow& cr = _coreRows[crIdx];
    	for(unsigned srIdx = 0; srIdx < cr.getNumSubRows(); ++srIdx)
    	{
      	    const RBPSubRow& sr = cr[srIdx];
      	    for(unsigned cOffset = 0; cOffset < sr.getNumCells(); cOffset++)
                _cellCoords[sr[cOffset]] = RBCellCoord(crIdx, srIdx, cOffset);
    	}
    }
}

void RBPlacement::setLocation(unsigned id, const Point& pt)
  //puts cell <id> at location <pt>. updates the row structures
{
    if(_isInSubRow[id])	//need to remove the cell from the sub-row it was in
    {
	RBCellCoord& coord = _cellCoords[id];
	RBPSubRow& sr = _coreRows[coord.coreRowId][coord.subRowId];
	sr.removeCell(id, coord);
	_isInSubRow[id] = false;
    }

    _placement[id] = pt;
    RBRowPtrs newRows = findBothRows(pt);
    if(newRows.first != NULL)
    {
	_cellCoords[id].coreRowId =newRows.first - &(*coreRowsBegin());
	_cellCoords[id].subRowId  =newRows.second-
		        &(*newRows.first->subRowsBegin());
	RBPSubRow* newSR = const_cast<RBPSubRow*>(newRows.second);
	for(unsigned idx = 0; idx < newSR->getNumCells(); idx++)	
	    if(_placement[(*newSR)[idx]].x >= pt.x) //insert here
	    {
		_cellCoords[id].cellOffset = idx;
		newSR->addCell(id, _cellCoords[id]);
	    }
	_isInSubRow[id] = true;
    }
}

void RBPlacement::sortSubRowsByX ()
{
  unsigned i,j;
  for(i=0; i<_coreRows.size(); i++) 
    for(j=0; j < _coreRows[i].getNumSubRows(); j++)
      _coreRows[i][j].sortCellsByLoc();
}

bool RBPlacement::isConsistent()
{
    bool looksOK = true;

    for(unsigned crId = 0; crId != _coreRows.size(); crId++)
    {
	RBPCoreRow& cr = _coreRows[crId];
	if(cr.getNumSubRows() == 0)
	{
	    looksOK = false;
	    cerr<<"CR: "<<crId<<endl;
	    cerr<<"CoreRow does not have any SubRows"<<endl;
	    continue;
	}

    	double prevXEnd = cr[0].getXStart();
    	for(unsigned srId = 0; srId != cr.getNumSubRows(); srId++)
        {
	    RBPSubRow& sr = cr[srId];
            if(sr._coreRow != &cr)
      	    {
		looksOK = false;
		cerr<<"CR: "<<crId<<" SR: "<<srId<<endl;
	    	cerr<<"SubRow does not point to the correct row"<<endl;
		if(sr._coreRow == NULL)
	    	    cerr<<"SubRow has NULL coreRow pointer"<<endl;
      	    }
      	    if(sr.getOrientation() != cr.getOrientation())
      	    {
        	looksOK = false;
        	cerr<<"Subrow orientation does not match its coreRow's"<<endl;
            }
            if(sr.getXStart() < prevXEnd)
            {
             	looksOK = false;
          	cerr<<"Mis-sorted x starts: "<<sr.getXStart()<<" "
			<<prevXEnd<< endl;
            }
	    if(sr.getNumSites() <= 0)
	    {
		looksOK = false;
		cerr<<"Empty SubRow"<<endl;
		cerr<<sr<<endl;
		cerr<<"Site width "<<sr.getSite().width<<endl;
	    }
	
	    for(unsigned cOffset = 0; cOffset != sr.getNumCells(); cOffset++)
	    {
		unsigned cellId = sr[cOffset];
		if(!_isInSubRow[cellId])
		{
		    looksOK = false;
		    cerr<<"subRow contains a cell marked as not in a SR"<<endl;
		}	
	 	else
		{
		    const RBCellCoord& cc = _cellCoords[cellId];
		    if( (cc.coreRowId != crId) ||
			(cc.subRowId  != srId) ||
			(cc.cellOffset != cOffset))
		    {
			looksOK = false;
			cerr<<"CellCoord does not match cells actual location";
			cerr<<endl;
		    }
		    if(_placement[cellId].y != sr.getYCoord())
		    {
			looksOK = false;
			cerr<<"Cell in subRow with mismatched y"<<endl;
		    }
		    if(_placement[cellId].x < sr.getXStart() ||
		       _placement[cellId].x > sr.getXEnd())
		    {
			looksOK = false;
			cerr<<"Cell in subRow with mismatched x"<<endl;
		    }
		}
	    }
   	}
    }

    return looksOK;
}

RBRowPtrs RBPlacement::findBothRows (const Point& point)
//   We assume that the RBPlacement contains _coreRows sorted by
//   y coordinate, and subrows sorted by x.  Return pointers
//   to the corerow and subrow that should contain the argument point.
{
  const RBPCoreRow* coreRowIdx = findCoreRow(point);

  if(coreRowIdx != NULL)
  {
    itRBPSubRow sr;
    for(sr = coreRowIdx->subRowsBegin(); sr != coreRowIdx->subRowsEnd(); sr++)
    {
      if(point.x >= sr->getXStart() && point.x <= sr->getXEnd())
      	return 
	  RBRowPtrs(std::pair<const RBPCoreRow*,const RBPSubRow*>
			   (coreRowIdx, &(*sr)));
    }

    return RBRowPtrs(coreRowIdx, static_cast<const RBPSubRow *>(NULL));
  }
  else
    return RBRowPtrs(static_cast<const RBPCoreRow *>(NULL),
                      static_cast<const RBPSubRow *>(NULL));
}

const RBPSubRow* RBPlacement::findSubRow (const Point& point) const
//   Return a pointer to the subrow containing <point>
{
  const RBPCoreRow* coreRowIdx = findCoreRow(point);

  if(coreRowIdx == NULL) return NULL;

  itRBPSubRow sr;
  for(sr = coreRowIdx->subRowsBegin(); sr != coreRowIdx->subRowsEnd(); sr++)
  {
    if(point.x >= sr->getXStart() && point.x <= sr->getXEnd())
	return &(*sr);
  }

  return NULL;	
}

const RBPCoreRow* RBPlacement::findCoreRow (const Point& point) const
  // Return a pointer to the _coreRow that should 
  // contain the argument point.
{
    RBPCoreRow cr;
    pair<const RBPCoreRow*, const RBPCoreRow*> coreRowIdx;
  
    cr._y = point.y;
    coreRowIdx = std::equal_range(&(*_coreRows.begin()), &(*_coreRows.end()), 
				cr, compareCoreRowY());

    if (coreRowIdx.first < coreRowIdx.second)
    	return coreRowIdx.first;
    else
    	return NULL;
}


void RBPlacement::extractCellsFromSR (RBPSubRow&        sr, 
				      vector<unsigned>& cellIds,
                                      vector<double>&   offsets,
                                      double beginOffset,  double endOffset)
{
    cellIds.erase(cellIds.begin(), cellIds.end());
    offsets.erase(offsets.begin(), offsets.end());
    cellIds.reserve(sr.getNumCells());
    offsets.reserve(sr.getNumCells());

    std::vector<unsigned>::iterator NULLit
	      = static_cast<vector<unsigned>::iterator>(NULL);
    std::vector<unsigned>::iterator eraseBegin = NULLit;
    std::vector<unsigned>::iterator eraseEnd   = NULLit;

    double epsilon = 0.01;

    beginOffset = max(0.0, beginOffset);
    endOffset   = min(sr.getLength() + epsilon, endOffset);

    double rangeBegin = sr.getXStart() + beginOffset;
    double rangeEnd   = sr.getXStart() + endOffset;

    for(unsigned i = 0; i < sr.getNumCells(); i++)
    {
	unsigned cellId = sr._cellIds[i];
	abkassert(_isInSubRow[cellId], "cell not marked as being in a sr");
        double  placedLoc = _placement[cellId].x;

	if(placedLoc < rangeBegin)
	{
	    abkassert(eraseBegin == NULLit, 
			"cell left of range after start found");
	    continue;
	}
	else if(placedLoc >= rangeEnd)
	{
	    if(eraseBegin == NULLit)	//all are right of the range
		return;
	    else if(eraseEnd == NULLit) //first one right of the range
		eraseEnd = sr._cellIds.begin() + i;

	    break;
	}
	else	//in the range..remove it
	{
	  if(eraseBegin == NULLit) eraseBegin = sr._cellIds.begin() + i;
	  abkassert(eraseEnd == NULLit,"sorting order of CellIds is incorrect");

	  cellIds.push_back(cellId);
	  offsets.push_back(placedLoc-sr.getXStart());

	  _cellsNotInRows++;
	  _isInSubRow[cellId] = false;
	  //maintain _CellCoords??
	}
    }
    if(eraseBegin == NULLit) return;
    if(eraseEnd   == NULLit) eraseEnd = sr._cellIds.end();

    sr._cellIds.erase(eraseBegin, eraseEnd);
}

void RBPlacement::extractCellsFromSR(
		 	RBPSubRow& 	  sr, 
			vector<unsigned>& cellIds,
                        vector<unsigned>& siteOffsets,
			unsigned beginSite, unsigned endSite)
{
    vector<double> offsets;
    extractCellsFromSR(sr, cellIds, offsets, beginSite*sr.getSpacing(),
						endSite*sr.getSpacing());
  
    siteOffsets.erase(siteOffsets.begin(), siteOffsets.end());
    siteOffsets.insert(siteOffsets.begin(), offsets.size(), 0);
    for(unsigned i = 0; i < offsets.size(); i++)
	siteOffsets[i] = static_cast<unsigned>(rint(offsets[i]/sr.getSpacing()));
}

void RBPlacement::extractCellsFromSR (RBPSubRow&        sr, 
				      vector<unsigned>& cellIds,
                                      double beginOffset,  double endOffset)
{
    cellIds.erase(cellIds.begin(), cellIds.end());
    cellIds.reserve(sr.getNumCells());

    std::vector<unsigned>::iterator NULLit
	      = static_cast<vector<unsigned>::iterator>(NULL);
    std::vector<unsigned>::iterator eraseBegin = NULLit;
    std::vector<unsigned>::iterator eraseEnd   = NULLit;

    double epsilon = 0.01;

    beginOffset = max(0.0, beginOffset);
    endOffset   = min(sr.getLength() + epsilon, endOffset);

    double rangeBegin = sr.getXStart() + beginOffset;
    double rangeEnd   = sr.getXStart() + endOffset;

    for(unsigned i = 0; i < sr.getNumCells(); i++)
    {
	unsigned cellId = sr._cellIds[i];
	abkassert(_isInSubRow[cellId], "cell not marked as being in a sr");
        double  placedLoc = _placement[cellId].x;

	if(placedLoc < rangeBegin)
	{
	    abkfatal(eraseBegin == NULLit, 
			"cell left of range after start found");
	    continue;
	}
	else if(placedLoc >= rangeEnd)
	{
	    if(eraseBegin == NULLit)	//all are right of the range
		return;
	    else if(eraseEnd == NULLit) //first one right of the range
		eraseEnd = sr._cellIds.begin() + i;

	    break;
	}
	else	//in the range..remove it
	{
	  if(eraseBegin == NULLit) eraseBegin = sr._cellIds.begin() + i;
	  abkassert(eraseEnd == NULLit, "sorting order of CellIds is incorrect");

	  cellIds.push_back(cellId);

	  _cellsNotInRows++;
	  _isInSubRow[cellId] = false;
	  //maintain _CellCoords??
	}
    }
    if(eraseBegin == NULLit) return;
    if(eraseEnd   == NULLit) eraseEnd = sr._cellIds.end();

    sr._cellIds.erase(eraseBegin, eraseEnd);
}


void RBPlacement::embedCellsInSR  (RBPSubRow& 	sr, 
				   const vector<unsigned>& cellIds,
                                   const vector<double>& spacing)
{
    abkfatal(spacing.size() == cellIds.size(), 
	"cellIds and spacing do not have the same number of entries");

    unsigned i = 0;
    for(unsigned newCell = 0; newCell < cellIds.size(); newCell++)
    {
	unsigned cellId      = cellIds[newCell];
	double newCellOffset = spacing[newCell];
		//find where cellId goes...

	for( ; i < sr._cellIds.size(); i++)
	  if(_placement[sr._cellIds[i]].x-sr.getXStart() > newCellOffset)
	 	break; 

	sr._cellIds.insert(sr._cellIds.begin()+i, cellId);

	double  loc = sr.getXStart() + newCellOffset;
	_placement[cellId].x = loc;
	_placement[cellId].y = sr.getYCoord();

	_cellsNotInRows--;
	_isInSubRow[cellId] = true;
	//maintain _CellCoords??
    }
}

void RBPlacement::embedCellsInSR   (RBPSubRow& 	sr, 
				    const vector<unsigned>& cellIds,
                                    const vector<unsigned>& siteSpacing)
{
    abkfatal(siteSpacing.size() == cellIds.size(), 
	"cellIds and spacing do not have the same number of entries");

    unsigned i = 0;
    for(unsigned newCell = 0; newCell < cellIds.size(); newCell++)
    {
	unsigned cellId = sr._cellIds[newCell];
	double newCellOffset = siteSpacing[newCell]*sr.getSpacing();
		//find where cellId goes...

	for( ; i < sr._cellIds.size(); i++)
	  if(_placement[sr._cellIds[i]].x-sr.getXStart() > newCellOffset)
	 	break; 

	sr._cellIds.insert(sr._cellIds.begin()+i, cellId);

	double  loc = sr.getXStart() + newCellOffset;
	_placement[cellId].x = loc;
	_placement[cellId].y = sr.getYCoord();

	_cellsNotInRows--;
	_isInSubRow[cellId] = true;
	//maintain _CellCoords??
    }
}

void RBPlacement::embedCellsInSR   (RBPSubRow& 	sr, 
				    const vector<unsigned>& cellIds,
                                    unsigned beginSite, unsigned endSite)
{
    if(cellIds.size() == 0) return;

    sr._cellIds.reserve(sr.getNumCells() + cellIds.size());

    beginSite = max((unsigned)0, beginSite);
    endSite   = min(sr.getNumSites(), endSite);

    double   rangeBegin = sr.getXStart() + sr.getSpacing()*beginSite;
    double   rangeEnd   = sr.getXStart() + sr.getSpacing()*endSite;

    cout<<"embedding range ["<<rangeBegin<<","<<rangeEnd<<")"<<endl;

    double totalCellWidth = 0;
    unsigned i;
    for(i = 0; i < cellIds.size(); i++)
    {
	double cellWidth = _hgWDims->getNodeWidth(cellIds[i]);
	totalCellWidth += sr.getCoreRow().siteRoundUp(cellWidth);
    }

    double spaceBetween = ((rangeEnd-rangeBegin)-totalCellWidth)/cellIds.size();

    cout<<"Total Site Width:      "<<rangeEnd-rangeBegin<<endl;
    cout<<"Total Cell Width:      "<<totalCellWidth<<endl;
    cout<<" ->Space Between Each: "<<spaceBetween<<endl<<endl;


    double leftEdge = rangeBegin;
    double cumExtraSpace = 0;	//extra space is accumulated as cells are
				//placed. when the extra space is > k*sitewidth,
				//k sites are skipped (for k any non-zero
				//integer)
    unsigned insPoint = 0;

    for(i = 0; i < cellIds.size(); i++)
    {
        unsigned newCellId = cellIds[i];

	for( ; insPoint < sr._cellIds.size(); insPoint++)
    	{
	  if(_placement[sr._cellIds[insPoint]].x > leftEdge)
		break;
	}

	sr._cellIds.insert(sr._cellIds.begin()+insPoint, newCellId);
	_placement[newCellId].x = leftEdge;
	_placement[newCellId].y = sr.getYCoord();

	_cellsNotInRows--;
	_isInSubRow[newCellId] = true;
	//maintain _CellCoords??

	double cellWidth = _hgWDims->getNodeWidth(newCellId);
	leftEdge += sr.getCoreRow().siteRoundUp(cellWidth);

	cumExtraSpace += spaceBetween;
	
	int numSitesOverlap = 
		static_cast<unsigned>(floor(cumExtraSpace/sr.getSpacing()));
	leftEdge += sr.getSpacing() * numSitesOverlap;
	cumExtraSpace -= sr.getSpacing() * numSitesOverlap;
    }
}

void RBPlacement::permuteCellsInSR(
			RBPSubRow& sr, 
			unsigned firstCellOffset,
                        const vector<unsigned>& cellPermutation)
{
abkassert(sr.getNumCells()-firstCellOffset >= cellPermutation.size(),
        "permuteCellsInSR has too many cells in the new ordering");

     std::vector<unsigned>::iterator permBegin = sr._cellIds.begin()+firstCellOffset;
     unsigned numToPermute = cellPermutation.size();

     double deltaSpace = sr.getSiteWidth(); //meaningless distance,
					//as a spacing function will be 
					//called after permuting

     double leftEdge = _placement[*permBegin].x;

     vector<unsigned> newCellOrder(numToPermute);
     unsigned k;
     for(k=0; k<numToPermute; k++)
        newCellOrder[k]=*(permBegin+cellPermutation[k]);

     for(k=0; k<numToPermute; k++)
     {
        unsigned newCellId = newCellOrder[k];
        *(permBegin+k) = newCellId;
        _placement[newCellId].x = leftEdge;

        leftEdge += deltaSpace;
     }

    switch(_params.spaceCellsAlg)
    {
        case RBPlacement::Parameters::EQUAL_SPACE:
            spaceCellsEqually(sr, firstCellOffset, numToPermute);
            break;
        case RBPlacement::Parameters::WITH_PIN_ALG1:
            spaceCellsWithPinDimensionsAlg1(sr, firstCellOffset, numToPermute);
            break;
        case RBPlacement::Parameters::WITH_PIN_ALG2:
            spaceCellsWithPinDimensionsAlg2(sr, firstCellOffset, numToPermute);
            break;
	default:
	    abkfatal(0,"unknown spacecells value");
    }
}

ostream& operator<<(ostream& out, const RBPlacement& arg)
{
  unsigned i,j;
  unsigned cellCount = 0;
  
  out << "RBPlacement object\n";
  out << static_cast<const Placement>(static_cast<const PlacementWOrient>(arg));
  out << "  " << arg._coreRows.size() << " core rows: ";
  for (i=0; i<arg._coreRows.size(); i++)
  {
    out << endl << "   " << arg._coreRows[i];
    for (j=0; j<arg._coreRows[i].getNumSubRows(); j++)
      cellCount += arg._coreRows[i][j].getNumCells();
  }
  out << endl;
  out << "   containing " << cellCount << " cell Id(s)" << endl;
  out << endl;
  return out;
}

RBPlacement::Parameters::Parameters(int argc, const char *argv[])
  :verb(argc,argv), numRowsToRemove(0), spaceCellsAlg(EQUAL_SPACE),
   remCongestion(0)
{
   UnsignedParam numRowsToRemove_("numRowsToRemove", argc, argv);
   StringParam spaceCellsAlg_("spaceCellsAlg", argc, argv);
   BoolParam remCongestion_("remCongestion", argc, argv);

   if(remCongestion_.found())
     remCongestion = 1;

   if(numRowsToRemove_.found())
	numRowsToRemove = numRowsToRemove_; 

   if(spaceCellsAlg_.found())
   {
      if(strcasecmp(spaceCellsAlg_,"EQUAL") == 0)
         spaceCellsAlg = EQUAL_SPACE;
      else if(strcasecmp(spaceCellsAlg_,"PIN1") == 0)
         spaceCellsAlg = WITH_PIN_ALG1;
      else if(strcasecmp(spaceCellsAlg_,"PIN2") == 0)
         spaceCellsAlg = WITH_PIN_ALG2;
      else
          abkwarn(0,"Unrecognize spaceCellsAlg! Using default value");
   }
}

ostream& operator<< (ostream& out, const RBPlaceParams& params)
{
    out<<" RBPlacement Parameters"<<endl;
    out<<"  Verbosity:         "<<params.verb<<endl;
    out<<"  NumRowsToRemove:   "<<params.numRowsToRemove<<endl;
    out<<"  SpaceCellsAlg:     ";
    switch(params.spaceCellsAlg)
    {
        case RBPlacement::Parameters::EQUAL_SPACE:
            out<<"Equal Spacing\n"; break;
        case RBPlacement::Parameters::WITH_PIN_ALG1:
            out<<"With Pin 1st algorithm\n"; break;
        case RBPlacement::Parameters::WITH_PIN_ALG2:
            out<<"With Pin 2nd algorithm\n"; break;
	default: out<<"Unknown Spacing Alg\n"; break;
    }
    out<<"  Remove Congestion: "<<params.remCongestion<<endl;

    return out;
}

void RBPlacement::savePlacement (const char* plFileName)   const
{
   ofstream out(plFileName); 
   
   out << "UCLA pl 1.0"<<endl;
   out <<TimeStamp() << User() << Platform() << endl << endl;

   for (unsigned i=0; i<_placement.getSize(); i++)
   {
        const char * namePtr=_hgWDims -> getNodeByIdx(i). getName();

        out<<setw(8) << (namePtr ? namePtr : "name unavailable")
                  <<"  "<<setw(10)<<_placement[i].x
                  <<"  "<<setw(10)<<_placement[i].y<<" : "
                  <<_placement.getOrient(i)<<endl;
   }

   out.close();          
}

void RBPlacement::savePlacementUnShredHW(const char* plFileName)   const
{
  ofstream out(plFileName); 
  
  out << "UCLA pl 1.0"<<endl;
  out <<TimeStamp() << User() << Platform() << endl << endl;

  unsigned i=0;
  for(i=0; i<_hgWDims->getNumNodes(); i++)
    {
      if (_placement[i].x > 1e308 || _placement[i].x > 1e308) continue;

      HGFNode& node   = _hgWDims->getNodeByIdx(i);
      const char* nodeName  = node.getName();
      char* actualName = new char[strlen(nodeName)+10];
      strcpy(actualName, nodeName);
      char* pos = strchr(actualName, '@');
      char* pos1 = strchr(actualName, '#');


      if(pos != NULL && pos1 != NULL)  //found
	{
	  *pos = '\0';
	  char *subNodeName = new char[strlen(nodeName)+10];
	  
	  double avgYLoc = 0;
	  double avgXLoc = 0;
	  unsigned numSubNodes = 0;
	  unsigned subVertIndex=0;
	  unsigned subHorizIndex=0;
	  char* indices = new char[strlen(nodeName)+10];

	  double nodeHeight = 0;
	  double nodeWidth = 0;
	  //Orientation nodeOrient = _placement.getOrient(i);

	  vector< vector<unsigned> > subNodes;
	  vector<unsigned> horizSubNodes;
	  unsigned lastVertIdx=0;
	  unsigned currVertIdx=0;

	  while(i<_hgWDims->getNumNodes())
	    {
	      HGFNode& subNode   = _hgWDims->getNodeByIdx(i);
	      nodeName = subNode.getName();
	      strcpy(subNodeName, nodeName);
	      pos = strchr(subNodeName, '@');
	      if(pos == NULL)
		{
		  subNodes.push_back(horizSubNodes);
		  horizSubNodes.clear();
		  break;
		}
	      *pos = '\0';
	      if(strcmp(actualName, subNodeName) != 0)
		{
		  subNodes.push_back(horizSubNodes);
		  horizSubNodes.clear();
		  break;
		}

	      strcpy(subNodeName, nodeName);
	      pos = strchr(subNodeName, '@');
	      *pos = '\0';
	      strcpy(indices, pos+1);
	      pos = strchr(indices, '#');
	      *pos = '\0';
	      subVertIndex = atoi(indices);
	      subHorizIndex = atoi(pos+1);
	      
	      currVertIdx = subVertIndex;
	      if(currVertIdx == lastVertIdx)
		horizSubNodes.push_back(i);
	      else
		{
		  subNodes.push_back(horizSubNodes);
		  horizSubNodes.clear();
		  horizSubNodes.push_back(i);
		  lastVertIdx = currVertIdx;
		}

	      if(subHorizIndex == 0)
		nodeHeight += _hgWDims->getNodeHeight(i);
	      if(subVertIndex == 0)
		nodeWidth += _hgWDims->getNodeWidth(i);

	      avgYLoc += (_placement[i].y + _hgWDims->getNodeHeight(i)/2);
	      avgXLoc += (_placement[i].x + _hgWDims->getNodeWidth(i)/2);
	      numSubNodes++;
	      i++;
	    }
	  if(horizSubNodes.size()!=0)
	    {
	      subNodes.push_back(horizSubNodes);
	      horizSubNodes.clear();
	    }
	  i--;
	  avgXLoc /= numSubNodes;
	  avgYLoc /= numSubNodes;

	  vector<int> orientScore(9,0);
	  unsigned k;
	  for(k=0; k<subNodes.size()-1; ++k)
	    {
	      for(unsigned l=0; l<subNodes[k].size()-1; ++l)
		{
		  clusterNodeOrient orient = 
		    getOrientSubNode(subNodes[k][l], 
				     subNodes[k][l+1], subNodes[k+1][l]);
		  switch(orient)
		    {
		    case N:
		      ++orientScore[N];
		      break;
		    case FN:
		      ++orientScore[FN];
		      break;
		    case S:
		      ++orientScore[S];
		      break;
		    case FS:
		      ++orientScore[FS];
		      break;
		    case W:
		      ++orientScore[W];
		      break;
		    case FW:
		      ++orientScore[FW];
		      break;
		    case E:
		      ++orientScore[E];
		      break;
		    case FE:
		      ++orientScore[FE];
		      break;
		    case I:
		      ++orientScore[I];
		      break;
		    default:
		      cout<<"Error in computing orientation "<<orient<<endl;
		    };
		}
	    }
	  
	  cout<<actualName<<endl;
	  for(k=0;k<9;++k)
	    cout<<setw(3)<<toChar(clusterNodeOrient(k));
	  cout<<endl;
	  
	  int maxScore=-1;
	  int maxScoreOrient=0;
	  for(unsigned k=0;k<8;++k)
	    {
	      if(orientScore[k] > maxScore)
		{
		  maxScore = orientScore[k];
		  maxScoreOrient = k;
		}
	      cout<<orientScore[k]<<" ";
	    }

	  //maxScoreOrient = N;

	  double nodeYLoc;
	  double nodeXLoc;
	  if(maxScoreOrient == N || maxScoreOrient == S || 
	     maxScoreOrient == FN || maxScoreOrient == FS)
	    {
	      nodeXLoc = avgXLoc - nodeWidth/2;
	      nodeYLoc = avgYLoc - nodeHeight/2;
	    }
	  else
	    {
	      nodeXLoc = avgXLoc - nodeHeight/2;
	      nodeYLoc = avgYLoc - nodeWidth/2;
	    }

	  out<<actualName<<"  ";
	  out<< setw(10) << nodeXLoc << "  ";
	  out<< setw(10) << nodeYLoc << " :  ";

	  cout<<orientScore[8];
	  cout<<endl;
	  out<<toChar(clusterNodeOrient(maxScoreOrient))<<endl;
	}
      else
	{
	  out << _hgWDims->getNodeByIdx(i).getName() <<"  "
		<< setw(10) << _placement[i].x << "  "
		<< setw(10) << _placement[i].y << " : "
		<< _placement.getOrient(i)<<endl;
	}
    }
    out.close();
}

RBPlacement::~RBPlacement()
{
  if(_hgWDims) delete _hgWDims;
  unsigned k;
  for(k=0; k!=allFileNamesInAUX.size(); k++) free(allFileNamesInAUX[k]);
}


void RBPlacement::permuteCellsAndWSInSR(
			RBPSubRow& sr, 
			unsigned firstCellOffset,
                        const vector<unsigned>& cellPermutation)
{

     std::vector<unsigned>::iterator permBegin = sr._cellIds.begin()+firstCellOffset;
     unsigned numToPermute = cellPermutation.size();
     vector<unsigned> movables;
     vector<unsigned> WSpositions;
     vector<double> whiteSpaceWidths;
     unsigned totalMovables=0;
     unsigned nodeCtr=firstCellOffset;
     unsigned totalCells = sr.getNumCells();

     int numSitesWS=0;
     int sitesPerWS=0;
     unsigned i;
     for(i=firstCellOffset;i<firstCellOffset+numToPermute;++i)
       {
	 if(nodeCtr==totalCells-1)   //last cell of the subRow
	   {
	     movables.push_back(*(sr.cellIdsBegin()+nodeCtr));
	     ++totalMovables;
	     ++nodeCtr;
	     if(totalMovables == numToPermute)
	       break;	     

	     //if whitespace present after last cell
	     double currWSWidth=sr.getXEnd() - 
	       _placement[*(sr.cellIdsBegin()+nodeCtr-1)].x-
	       _hgWDims->getNodeWidth(*(sr.cellIdsBegin()+nodeCtr-1));
	     
	     if(currWSWidth >= 2*sr.getSpacing())
	       {
		 numSitesWS = int(currWSWidth/sr.getSpacing());

		 movables.push_back(UINT_MAX);
		 sitesPerWS = int(floor(numSitesWS/2+0.5));
		 numSitesWS -= sitesPerWS;
		 whiteSpaceWidths.push_back(sitesPerWS*sr.getSpacing());
		 WSpositions.push_back(totalMovables);
		 ++totalMovables;
		 if(totalMovables == numToPermute)
		   break;

		 movables.push_back(UINT_MAX);
		 sitesPerWS = numSitesWS;
		 numSitesWS -= sitesPerWS;
		 whiteSpaceWidths.push_back(sitesPerWS*sr.getSpacing());
		 WSpositions.push_back(totalMovables);
		 ++totalMovables;
		 if(totalMovables == numToPermute)
		   break;
	       }
	     else
	       {
		 movables.push_back(UINT_MAX);
		 whiteSpaceWidths.push_back(currWSWidth);
		 WSpositions.push_back(totalMovables);
		 ++totalMovables;
		 if(totalMovables == numToPermute)
		   break;		
	       }
	   }
	 else
	   {
	     //if whitespace found
	     if(_placement[*(sr.cellIdsBegin()+nodeCtr)].x+
		_hgWDims->getNodeWidth(*(sr.cellIdsBegin()+nodeCtr)) <
		_placement[*(sr.cellIdsBegin()+nodeCtr+1)].x)
	       {
		 movables.push_back(*(sr.cellIdsBegin()+nodeCtr));
		 ++totalMovables;
		 ++nodeCtr;
		 if(totalMovables == numToPermute)
		   break;

		 //better white space handling
		 double currWSWidth=_placement[*(sr.cellIdsBegin()+nodeCtr)].x-
		   _placement[*(sr.cellIdsBegin()+nodeCtr-1)].x-
		   _hgWDims->getNodeWidth(*(sr.cellIdsBegin()+nodeCtr-1));

		 
		 if(currWSWidth >= 3*sr.getSpacing())
		   {
		     numSitesWS = int(currWSWidth/sr.getSpacing());

		     movables.push_back(UINT_MAX);
		     sitesPerWS = int(floor(numSitesWS/3+0.5));
		     numSitesWS -= sitesPerWS;
		     whiteSpaceWidths.push_back(sitesPerWS*sr.getSpacing());
		     WSpositions.push_back(totalMovables);
		     ++totalMovables;
		     if(totalMovables == numToPermute)
		       break;

		     movables.push_back(UINT_MAX);
		     sitesPerWS = int(floor(numSitesWS/2+0.5));
		     numSitesWS -= sitesPerWS;
		     whiteSpaceWidths.push_back(sitesPerWS*sr.getSpacing());
		     WSpositions.push_back(totalMovables);
		     ++totalMovables;
		     if(totalMovables == numToPermute)
		       break;

		     movables.push_back(UINT_MAX);
		     sitesPerWS = numSitesWS;
		     numSitesWS -= sitesPerWS;
		     whiteSpaceWidths.push_back(sitesPerWS*sr.getSpacing());
		     WSpositions.push_back(totalMovables);
		     ++totalMovables;
		     if(totalMovables == numToPermute)
		       break;
		   }
		 else if(currWSWidth < 3*sr.getSpacing() && 
			 currWSWidth >=2*sr.getSpacing())
		 
		 //if(currWSWidth >=2*sr.getSpacing())
		   {
		     movables.push_back(UINT_MAX);
		     whiteSpaceWidths.push_back(currWSWidth/2);
		     WSpositions.push_back(totalMovables);
		     ++totalMovables;
		     if(totalMovables == numToPermute)
		       break;
		     movables.push_back(UINT_MAX);
		     whiteSpaceWidths.push_back(currWSWidth/2);
		     WSpositions.push_back(totalMovables);
		     ++totalMovables;
		     if(totalMovables == numToPermute)
		       break;
		   }
		 else
		 
		   {
		     movables.push_back(UINT_MAX);
		     whiteSpaceWidths.push_back(currWSWidth);
		     WSpositions.push_back(totalMovables);
		     ++totalMovables;
		     if(totalMovables == numToPermute)
		       break;
		   }
	       }
	     else
	       {
		 movables.push_back(*(sr.cellIdsBegin()+nodeCtr));
		 ++totalMovables;
		 ++nodeCtr;
		 if(totalMovables == numToPermute)
		   break;

	       }
	   }
       }

    abkassert(movables.size() == cellPermutation.size(),
      "permuteCellsInSR has too many cells in the new ordering");

     for(i=0;i<WSpositions.size();++i)
       {
	 for(unsigned j=0;j<numToPermute;++j)
	   {
	     if(WSpositions[i]==cellPermutation[j])
	       {
		 WSpositions[i]=j;
		 break;
	       }
	   }
       }

     double leftEdge=_placement[movables[0]].x;
     nodeCtr=0;
     for(unsigned i=0;i<numToPermute;++i)
       {
	 if(movables[cellPermutation[i]] != UINT_MAX)
	   {
	     _placement[movables[cellPermutation[i]]].x=leftEdge;
	     leftEdge+=_hgWDims->getNodeWidth(movables[cellPermutation[i]]);
	     *(permBegin+nodeCtr) = movables[cellPermutation[i]];
	     ++nodeCtr;
	   }
	 else
	   {
	     for(unsigned int j=0;j<WSpositions.size();++j)
	       {
		 if(i==WSpositions[j])
		   { 
		     leftEdge+=whiteSpaceWidths[j];
		     break;
		   }
	       }
	   }
       }
}

//added by sadya to update placement of ironed cells

void RBPlacement::updateIronedCells(const vector<unsigned>& movables,
				    const Placement& soln,
				    RBPSubRow& subrow1,
				    RBPSubRow& subrow2)
{
  vector<unsigned> row1Ids;
  vector<unsigned> row2Ids;
  double yCoord1=subrow1.getYCoord();
  double yCoord2=subrow2.getYCoord();

  unsigned i;
  for(i=0;i<movables.size();++i)
    {
      if(movables[i] != UINT_MAX)
	{
	  if(_placement[movables[i]].y == yCoord1)
	    {
	      subrow1.removeCell(movables[i]);
	      _isInSubRow[movables[i]] = false;
	    }
	  else if(_placement[movables[i]].y == yCoord2)
	    {
	      subrow2.removeCell(movables[i]);
	      _isInSubRow[movables[i]] = false;
	    }
	  else
	    abkfatal(0,"invalid small placement problem\n");
	}
    }
  for(i = 0; i < movables.size(); i++)
    {
      if(movables[i] != UINT_MAX)
	{
	  _placement[movables[i]] = Point(soln[i].x, soln[i].y);

	  if(soln[i].y == yCoord1)
	    {
	      subrow1.addCell(movables[i]);
	      _isInSubRow[movables[i]] = true;
	    }
	  else if(soln[i].y == yCoord2)
	    {
	      subrow2.addCell(movables[i]);
	      _isInSubRow[movables[i]] = true;
	    }
	  else
	    abkfatal(0,"invalid small placement solution\n");
	} 
    }
}



//added by sadya to update placement of ironed cells with clustering

void RBPlacement::updateIronedCellsLR(const vector< vector<unsigned> >& movables,
				    const Placement& soln,
				    RBPSubRow& subrow1,
				    RBPSubRow& subrow2)
{
  vector<unsigned> row1Ids;
  vector<unsigned> row2Ids;
  double yCoord1=subrow1.getYCoord();
  double yCoord2=subrow2.getYCoord();

  unsigned i;
  for(i=0;i<movables.size();++i)
    {
      if(movables[i][0] != UINT_MAX)
	{
	  for(unsigned k=0;k<movables[i].size();++k)
	    {
	      if(_placement[movables[i][k]].y == yCoord1)
		{
		  subrow1.removeCell(movables[i][k]);
		  _isInSubRow[movables[i][k]] = false;
		}
	      else if(_placement[movables[i][k]].y == yCoord2)
		{
		  subrow2.removeCell(movables[i][k]);
		  _isInSubRow[movables[i][k]] = false;
		}
	      else
		abkfatal(0,"invalid small placement problem\n");
	    }
	}
    }
  for(i = 0; i < movables.size(); i++)
    {
      if(movables[i][0] != UINT_MAX)
	{
	  double leftEdge = soln[i].x;
	  for(unsigned k=0;k<movables[i].size();++k)
	    {
	      _placement[movables[i][k]] = Point(leftEdge, soln[i].y);
	      leftEdge += (*_hgWDims).getNodeWidth(movables[i][k]);

	      if(soln[i].y == yCoord1)
		{
		  subrow1.addCell(movables[i][k]);
		  _isInSubRow[movables[i][k]] = true;
		}
	      else if(soln[i].y == yCoord2)
		{
		  subrow2.addCell(movables[i][k]);
		  _isInSubRow[movables[i][k]] = true;
		}
	      else
		abkfatal(0,"invalid small placement solution\n");
	    } 
	}
    }
}


//added by sadya to update placement of ironed cells with clustering R->L

void RBPlacement::updateIronedCellsRL(const vector< vector<unsigned> >& movables,
				    const Placement& soln,
				    RBPSubRow& subrow1,
				    RBPSubRow& subrow2)
{
  vector<unsigned> row1Ids;
  vector<unsigned> row2Ids;
  double yCoord1=subrow1.getYCoord();
  double yCoord2=subrow2.getYCoord();

  unsigned i;
  for(i=0;i<movables.size();++i)
    {
      if(movables[i][0] != UINT_MAX)
	{
	  for(unsigned k=0;k<movables[i].size();++k)
	    {
	      if(_placement[movables[i][k]].y == yCoord1)
		{
		  subrow1.removeCell(movables[i][k]);
		  _isInSubRow[movables[i][k]] = false;
		}
	      else if(_placement[movables[i][k]].y == yCoord2)
		{
		  subrow2.removeCell(movables[i][k]);
		  _isInSubRow[movables[i][k]] = false;
		}
	      else
		abkfatal(0,"invalid small placement problem\n");
	    }
	}
    }


  for(i = 0; i < movables.size(); i++)
    {
      if(movables[i][0] != UINT_MAX)
	{
	  double leftEdge = soln[i].x;
	  for(unsigned k=0 ; k < movables[i].size();++k)
	    {
	      _placement[movables[i][k]] = Point(leftEdge, soln[i].y);

	      leftEdge += (*_hgWDims).getNodeWidth(movables[i][k]);

	      if(soln[i].y == yCoord1)
		{
		  subrow1.addCell(movables[i][k]);
		  _isInSubRow[movables[i][k]] = true;
		}
	      else if(soln[i].y == yCoord2)
		{
		  subrow2.addCell(movables[i][k]);
		  _isInSubRow[movables[i][k]] = true;
		}
	      else
		abkfatal(0,"invalid small placement solution\n");
	    } 
	}
    }
}

//added by sadya to get the BBox of the core layout
BBox RBPlacement::getBBox(bool addTerminals)
{
  BBox bbox;
  itRBPCoreRow itc;
  for(itc = coreRowsBegin(); itc != coreRowsEnd(); ++itc)
    {
      bbox.add(itc->getXStart(), itc->getYCoord());
      bbox.add(itc->getXEnd(), itc->getYCoord()+itc->getHeight());
    }

  //add to bbox all the fixed cells
  if(addTerminals)
    {
      for(unsigned i=0; i<_hgWDims->getNumNodes(); i++)
	{
	  if(isFixed(i))
	    {
	      bbox.add(_placement[i].x, _placement[i].y);
	      bbox.add(_placement[i].x + _hgWDims->getNodeWidth(i),
		       _placement[i].y + _hgWDims->getNodeHeight(i));
	      
	    }
	}
    }
  return bbox;
}

clusterNodeOrient RBPlacement::getOrientSubNode(unsigned index, 
						unsigned indexRight, 
						unsigned indexTop) const
{
  clusterNodeOrient orient=I;
  Point node(_placement[index]);
  Point nodeRight(_placement[indexRight]);
  Point nodeAbove(_placement[indexTop]);

  Point V2(nodeRight.x-node.x, nodeRight.y-node.y);
  Point V1(nodeAbove.x-node.x, nodeAbove.y-node.y);

  int quad1 = getQuad(V1);
  int quad2 = getQuad(V2);

  //cout<<"V1: "<<V1<<" quad1: "<<quad1<<"V2: "<<V2<<" quad2: "<<quad2<<endl;

  switch(quad2)
    {
    case 0:
      switch(quad1)
	{
	case 0:
	case 2:
	  break;
	case 1:
	  orient = N;
	  break;
	case 3:
	  orient = FS;
	  break;
	};
      break;

    case 1:
      switch(quad1)
	{
	case 1:
	case 3:
	  break;
	case 0:
	  orient = FW;
	  break;
	case 2:
	  orient = W;
	  break;
	};
      break;

    case 2:
      switch(quad1)
	{
	case 2:
	case 0:
	  break;
	case 1:
	  orient = FN;
	  break;
	case 3:
	  orient = S;
	  break;
	};      
      break;

    case 3:
      switch(quad1)
	{
	case 3:
	case 1:
	  break;
	case 0:
	  orient = E;
	  break;
	case 2:
	  orient = FE;
	  break;
	};
      break;
    };

  //cout<<"quad1: "<<quad1<<" quad2: "<<quad2<<" orient "<<toChar(orient)<<endl;
  /*
  if(_placement[index].x < _placement[indexRight].x &&
     _placement[index].y <= _placement[indexTop].y)
    orient = N;
  else if(_placement[index].x > _placement[indexRight].x &&
	  _placement[index].y <= _placement[indexTop].y)
    orient = FN;
  else if(_placement[index].x <= _placement[indexRight].x &&
	  _placement[index].y > _placement[indexTop].y)
    orient = S;
  else if(_placement[index].x >= _placement[indexRight].x &&
	  _placement[index].y > _placement[indexTop].y)
    orient = FS;
  else if(_placement[index].x >= _placement[indexTop].x &&
	  _placement[index].y <= _placement[indexRight].y)
    orient = W;
  else if(_placement[index].x <= _placement[indexTop].x &&
	  _placement[index].y <= _placement[indexRight].y)
    orient = FW;
  else if(_placement[index].x <= _placement[indexTop].x &&
	  _placement[index].y >= _placement[indexRight].y)
    orient = E;
  else if(_placement[index].x >= _placement[indexTop].x &&
	  _placement[index].y >= _placement[indexRight].y)
    orient = FE;
  else
    cout<<"ERROR"<<endl;
  */
  //orient = N;

  return orient;
}

int RBPlacement::getQuad(Point& pt) const
{
  if(pt.x > pt.y)
    {
      if(pt.x > -pt.y)
	return 0;
      else
	return 3;
    }
  else
    {
      if(pt.x > -pt.y)
	return 1;
      else
	return 2;
    }
}

char * RBPlacement::toChar(clusterNodeOrient orient) const
{
  switch(orient)
    {
    case N:
      return("N");
      break;
    case FN:
      return("FN");
      break;
    case S:
      return("S");
      break;
    case FS:
      return("FS");
      break;
    case W:
      return("W");
      break;
    case FW:
      return("FW");
      break;
    case E:
      return("E");
      break;
    case FE:
      return("FE");
      break;
    default:
      return("I");
      break;
    };
  cout<<orient<<" ERROR"<<endl;
  return("N");
}

void RBPlacement::removeSitesBelowMacros(void)
{
  double nodeWidth, nodeHeight, rowHeight;
  for(unsigned i=0; i<_hgWDims->getNumNodes(); i++)
    {
      //const HGFNode& node = _hgWDims->getNodeByIdx(i);
      Orient nodeOrient = _placement.getOrient(i);
      if(nodeOrient == Orient("N") || nodeOrient == Orient("S") || 
	 nodeOrient == Orient("FN") || nodeOrient == Orient("FS"))
	{
	  nodeWidth = _hgWDims->getNodeWidth(i);
	  nodeHeight = _hgWDims->getNodeHeight(i);
	}
      else
	{
	  nodeWidth = _hgWDims->getNodeHeight(i);
	  nodeHeight = _hgWDims->getNodeWidth(i);
	}

      BBox obstacle;
      obstacle.add(_placement[i].x, _placement[i].y);
      obstacle.add(_placement[i].x+nodeWidth, _placement[i].y+nodeHeight);

      RBRowPtrs rowPtrs = findBothRows(_placement[i]);

      if(rowPtrs.first != NULL && rowPtrs.second != NULL)
	rowHeight = (rowPtrs.first)->getHeight();
      else
	rowHeight = _coreRows[0].getHeight();

      if(nodeHeight > rowHeight && _hgWDims->isTerminal(i))
	{
	  for(unsigned i=0; i<_coreRows.size(); i++)
	    {
	      RBPCoreRow& cr = _coreRows[i];
	      double coreRowYMin = cr.getYCoord();
	      double coreRowYMax = cr.getYCoord() + cr.getHeight();
	      
	      if((obstacle.yMin > coreRowYMin) || 
		 (obstacle.yMax < coreRowYMax))  continue;
	      
	      splitRowByObstacle(cr, obstacle);
	    }
	}
    }
}

unsigned RBPlacement::getNumMacros(void)
{
  double nodeHeight, rowHeight = _coreRows[0].getHeight();
  unsigned numMacros = 0;
  for(unsigned i=0; i<_hgWDims->getNumNodes(); i++)
    {
      //const HGFNode& node = _hgWDims->getNodeByIdx(i);
      Orient nodeOrient = _placement.getOrient(i);
      if(nodeOrient == Orient("N") || nodeOrient == Orient("S") || 
	 nodeOrient == Orient("FN") || nodeOrient == Orient("FS"))
	nodeHeight = _hgWDims->getNodeHeight(i);
      else
	nodeHeight = _hgWDims->getNodeWidth(i);
      
      if(nodeHeight > rowHeight)
	++numMacros;
    }
  return numMacros;
}

void RBPlacement::updatePlacement(unsigned id, const Point& pt)
{
  _placement[id] = pt;
}

//To Do
void RBPlacement::assignPinsToCells()
{
  itHGFEdgeLocal edge;
  for(unsigned i=0; i<_hgWDims->getNumNodes(); i++)
    {
      const HGFNode& node = _hgWDims->getNodeByIdx(i);
      double nodeHeight = _hgWDims->getNodeHeight(i);
      double nodeWidth = _hgWDims->getNodeWidth(i);

      unsigned N = node.getDegree();
      unsigned numPinsTB = unsigned(ceil(N*nodeWidth/(2*(nodeWidth+nodeHeight))));
      unsigned numPinsLR = unsigned(ceil(N*nodeHeight/(2*(nodeWidth+nodeHeight))));
      //double xGridSize = nodeWidth/numPinsTB;
      //double yGridSize = nodeHeight/numPinsLR;
      vector<bool> left(numPinsLR,0);
      vector<bool> right(numPinsLR,0);
      vector<bool> top(numPinsTB,0);
      vector<bool> bottom(numPinsTB,0);
      
      unsigned e=0; //edge offset
      for(edge = node.edgesBegin(); edge != node.edgesEnd(); edge++, e++)
	{
	  //Point newPinOffset(nodeWidth/2, nodeHeight/2); //default to center
	  Point optLoc(0,0);
	  unsigned edgeIdx = (*edge)->getIndex();
	  unsigned nodeOffset = 0;
	  for(itHGFNodeLocal currNode = (*edge)->nodesBegin(); 
	      currNode != (*edge)->nodesEnd(); ++currNode, ++nodeOffset)
	    {
	      unsigned currNodeIdx = (*currNode)->getIndex();
	      Point currNodePOffset = _hgWDims->nodesOnEdgePinOffset
		(nodeOffset,edgeIdx,_placement.getOrient(currNodeIdx));
	      Point currNodePlacement = 
		_placement[currNodeIdx] + currNodePOffset;
	      optLoc += currNodePlacement;
	    }
	  
	  Point pinLoc = (_placement[i] + _hgWDims->edgesOnNodePinOffset
			  (e,i,_placement.getOrient(i)));  

	  optLoc -= pinLoc;
	  optLoc.x /= (nodeOffset-1);
	  optLoc.y /= (nodeOffset-1);
	  
	  Point newPinLoc;

	  if(optLoc.x <= _placement[i].x)
	    newPinLoc.x = _placement[i].x;
	  else if(optLoc.x >= _placement[i].x + nodeWidth)
	    newPinLoc.x = _placement[i].x + nodeWidth;

	  if(optLoc.y <= _placement[i].y)
	    newPinLoc.y = _placement[i].y;
	  else if(optLoc.y >= _placement[i].y + nodeHeight)
	    newPinLoc.y = _placement[i].y + nodeHeight;

	  cout<<optLoc.x<<" "<<optLoc.y<<endl;
	}
    }
}


